local ADDON_NAME="AVR"

local _G = _G
local pairs = pairs
local ipairs = ipairs
local remove = table.remove
local insert = table.insert
local type = type
local select = select
local next = next
local format = string.format

AVR = LibStub("AceAddon-3.0"):NewAddon(ADDON_NAME, "AceEvent-3.0", "AceConsole-3.0", "AceHook-3.0", "AceSerializer-3.0", "AceComm-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale(ADDON_NAME, true)

local Compress=LibStub("LibCompress")
local AddonEncodeTable=Compress:GetAddonEncodeTable()

local Core=AVR

local function tableFind(table,value)
	for i,v in ipairs(table) do
		if v==value then return i end
	end
	return nil
end

local function capitalize(s)
	return (s:sub(1,1):upper())..(s:sub(2):lower())
end

function Core:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New(ADDON_NAME.."_DB",Core.defaultOptions)
	
	Core.options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	Core.AceConfig=LibStub("AceConfig-3.0")
	Core.AceConfig:RegisterOptionsTable(ADDON_NAME, Core.options, {ADDON_NAME:lower()})
	Core.AceConfigRegistry=LibStub("AceConfigRegistry-3.0")
	Core.AceConfigDialog=LibStub("AceConfigDialog-3.0")
	Core.AceConfigDialog:AddToBlizOptions(ADDON_NAME, ADDON_NAME)

	self.db.RegisterCallback(self,"OnProfileShutdown","OnProfileShutdown")
	self.db.RegisterCallback(self,"OnDatabaseShutdown","OnProfileShutdown")
	self.db.RegisterCallback(self,"OnProfileChanged","OnProfileChanged")
	self.db.RegisterCallback(self,"OnProfileCopied","OnProfileChanged")
	self.db.RegisterCallback(self,"OnProfileReset","OnProfileChanged")
	
	self:RegisterComm(ADDON_NAME,"OnCommReceived")
end

function Core:OnCommReceived(prefix,message,distribution,sender)
	if not self.db.profile.enable then return end
	if self.db.profile.receiveOwnScenes=="no" and sender==GetUnitName("player") then return end
	
	if distribution=="RAID" then
		if not self.db.profile.acceptScenesFromRaid and not self.db.profile.acceptScenesFromRaidLeader and
			not self.db.profile.acceptScenesFromRaidAssist then return end
		local rank=0
		for i=1,40 do
			local raidName,raidRank=GetRaidRosterInfo(i)
			if raidName==sender then
				rank=raidRank
				break
			end
		end
		if rank==0 and not self.db.profile.acceptScenesFromRaid then return end
		if rank==1 and not self.db.profile.acceptScenesFromRaidAssist then return end
		if rank==2 and not self.db.profile.acceptScenesFromRaidLeader then return end
	elseif distribution=="PARTY" then
		if not self.db.profile.acceptScenesFromParty and not self.db.profile.acceptScenesFromPartyLeader then 
			return end
		local rank=0
		local leaderIndex=GetPartyLeaderIndex()
		if leaderIndex~=0 then
			for i=1,4 do
				local partyName=UnitName("party"..i)
				if partyName==sender then
					if leaderIndex==i then rank=2 end
					break
				end
			end
		elseif sender==GetUnitName("player") then rank=2
		end
		if rank==0 and not self.db.profile.acceptScenesFromParty then return end
		if rank==2 and not self.db.profile.acceptScenesFromPartyLeader then return end
	elseif distribution=="BATTLEGROUND" and not self.db.profile.acceptScenesFromBG then return 
	elseif distribution=="GUILD" and not self.db.profile.acceptScenesFromGuild then return 
	elseif distribution=="WHISPER" and not self.db.profile.acceptScenesFromWhisper then return
	end

	self:Print(format(L["Received AVR scene from %s"],sender))
	message=AddonEncodeTable:Decode(message)
	message=Compress:Decompress(message)
	local ok,msg=self:Deserialize(message)
	if not ok then 
		self:Print(format(L["Couldn't deserialized AVR addon message from %s %s"],sender,msg))
		return
	end
	if msg.type=="scene" then
		local scene=self.sceneManager:UnpackScene(msg.content)
		scene.from=sender
		if sender==GetUnitName("player") and self.db.profile.receiveOwnScenes=="hidden" then
			scene.visible=false
		end
		if not scene.id then scene.id=1 end
		self.sceneManager:ReceiveScene(scene)
	end
end

function Core:SendScene(scene,distribution,target,prio)
	if scene.from~=nil then
		self:Print(L["Can't send a scene from someone else. Make own copy first."])
		return
	end

	distribution=distribution or "RAID"
	prio=prio or "BULK"
	local msg={
		type="scene",
		content=scene:Pack()
	}
	msg=self:Serialize(msg)
	msg=Compress:CompressHuffman(msg)
	msg=AddonEncodeTable:Encode(msg)
	self:Print(format(L["Sending scene to %s"],distribution=="WHISPER" and target or L[distribution:lower()]))
	self:SendCommMessage(ADDON_NAME,msg,distribution,target,prio)
end

function Core:OnProfileShutdown()
	if self.sceneManager then self.sceneManager:SaveScenes() end
end
function Core:OnProfileChanged()
	if self.sceneManager then self.sceneManager:LoadScenes() end
end

function Core.Embed(self, other)
	local k,v
	if _G.getmetatable(other) then
		for k, v in pairs(self) do
			other[k] = v
		end
	else
		_G.setmetatable(other, {__index = self})
	end
	return other
end

function Core:EnableDraw()
	self.rootFrame:EnableMouse(true)
end

function Core:DisableDraw()
	self.rootFrame:EnableMouse(false)
end

function Core:OnUpdate()
	self.zoneData:OnUpdate()	
	self.threed:StartRender()
	self.mousePaint:UpdatePaint()
	if not self.db.profile.hideAll then
		self.sceneManager:DrawScenes()
	end
	self.threed:EndRender()
end

function Core:OnEnable()
	self:DoEnable()
end
function Core:DoEnable()
	if not self.db.profile.enable then return end
	if not self.rootFrame then
		self.rootFrame=CreateFrame("FRAME",ADDON_NAME.."_RootFrame",WorldFrame)
		self.rootFrame:SetPoint("BOTTOMLEFT",WorldFrame,"BOTTOMLEFT",0,0)
		self.rootFrame:SetPoint("TOPRIGHT",WorldFrame,"TOPRIGHT",0,0)
	end
	if not self.zoneData then
		self.zoneData=AVRZoneData:New()
	end
	self.zoneData:Enable()
	if not self.threed then
		self.threed=AVR3D:New(self.rootFrame,self)
	end
	if not self.sceneManager then
		self.sceneManager=AVRSceneManager:New(self.threed)
	end
	self.sceneManager:LoadScenes()
	if not self.mousePaint then
		self.mousePaint=AVRMousePaint:New(self.threed)
	end
	self.rootFrame:SetScript("OnUpdate",function() self:OnUpdate() end )
	
	self:OptionsChanged()
end


function Core:OnDisable()
	self:DoDisable()
end
function Core:DoDisable()

	self.rootFrame:SetScript("OnUpdate",nil )
	if self.mousePaint then
		self.mousePaint:DisableDraw()
		self.mousePaint:DisableDrag()
	end
	if self.threed then
		self.threed:ClearAllLines()
		self.threed:ClearAllTriangles()
	end	
	if self.zoneData then
		self.zoneData:Disable()
	end
	if self.sceneManager then
		self.sceneManager:SaveScenes()
		self.sceneManager=nil
	end
	self:UnregisterAllEvents()
	self:OptionsChanged()
end

function Core:GetTargetPlatePos()
	local cs={WorldFrame:GetChildren()}
	local c,r
	local t=nil
	for i=1,#cs do
		c=cs[i]
		r=c:GetRegions()
		if r and r:GetObjectType() == "Texture" and r:GetTexture() == "Interface\\TargetingFrame\\UI-TargetingFrame-Flash" then
			if c:GetAlpha()==1 then
				if t~=nil then 
					return nil
				else
					t=r
				end
			end
		end
	end
	if t==nil then
		return nil
	end
	return (t:GetLeft()+t:GetRight())/2,t:GetTop()
end

function Core:Benchmark(num)
	local x,y
	num=num or 5
	for x=-num,num do 
		for y=-num,num do
			self.scene:AddMesh(AVRCircle(0.5,20):TranslateMesh(x,y))
			self.scene:GetLastMesh():SetFollowPlayer(true):Detach()
		end
	end
end

function Core:GetSelectedMesh()
	if not self.sceneManager then return nil end
	local scene=self.sceneManager:GetSelectedScene()
	if not scene then return nil end
	return scene:GetSelectedMesh()
end

function Core:AddMesh(mesh)
	if not self.sceneManager then return nil end
	local scene=self.sceneManager:GetSelectedScene()
	if not scene then return nil end
	return scene:AddMesh(mesh)
end

function Core:FindTargetPlateHeight()
	local px,py=self.threed:GetUnitPosition("player")
	local x,y=self:GetTargetPlatePos()
	if x==nil then
		return nil
	end
	
	self.threed:InvertCameraMatrix()
	local px1,py1,pz1=self.threed:InverseProject(x,y,20)
	local px2,py2,pz2=self.threed:InverseProject(x,y,10)
	local dx1,dy1=px2-px1,py2-py1
	local dx2,dy2=px-px1,py-py1
	local d=(dx1*dx2+dy1*dy2)/(dx1*dx1+dy1*dy1)
	local h=pz1+(pz2-pz1)*d+self.threed.playerHeight
	return h
end
