local ADDON_NAME="AVR"
local L = LibStub("AceLocale-3.0"):GetLocale(ADDON_NAME, true)


local _G = _G
local pairs = pairs
local ipairs = ipairs
local remove = table.remove
local insert = table.insert
local max = math.max
local min = math.min
local abs = math.abs
local atan2 = math.atan2
local sin = math.sin
local cos = math.cos
local type = type
local select = select
local next = next
local GetTime = GetTime
local deg2rad = math.pi/180.0
local pi = math.pi
local format = string.format
local tostring,tonumber=tostring,tonumber
local strlen=string.len

local Core=AVR


local function packDouble(d)
	if d==nil then return d end
	if Core.db.profile.noPackDoubles then return d end
	local s=tostring(d)
	if tonumber(s)==d and strlen(s)<8 then return d end
	local bs=AVRByteStream:New()
	bs:WriteDouble(d)
	bs:FlushBuffer()
	return bs.data[1]
end
local function unpackDouble(p)
	if p==nil then return p end
	if type(p)=="number" then return p end
	local bs=AVRByteStream:New({data={p}})
	return bs:ReadDouble()
end
AVRUnpackDouble=unpackDouble
AVRPackDouble=packDouble

AVRMesh = {Embed=Core.Embed}
AVRMesh.class="AVRMesh"
function AVRMesh:New(scene)
	if self ~= AVRMesh then return end
	local s={}
	
	AVRMesh:Embed(s)
	
	s.name=""
	s.class=AVRMesh.class
	-- each vertex is an array of 8, first three are mesh coordinates, next three
	-- coordinates after camera rotation, last two projected screen coordinates
	s.vertices={}
	s.lines={}
	s.triangles={}
	s.meshTranslateX=0
	s.meshTranslateY=0
	s.meshTranslateZ=0
	s.meshRotateZ=0
	s.meshScaleX=1.0
	s.meshScaleY=1.0
	s.meshScaleZ=1.0
	s.translateX=0
	s.translateY=0
	s.translateZ=0
	s.rotateZ=0
	s.scaleX=1.0
	s.scaleY=1.0
	s.scaleZ=1.0
	s.visible=true
	s.r=1.0
	s.g=1.0
	s.b=1.0
	s.a=1.0
	
	s.followUnit=nil
	s.followPlayer=false
	s.followRotation=false
	
	s.removed=false
	
	s:SetScene(scene)

	return s
end

function AVRMesh:GetOptions()
	return {
		name = self.name,
		type = "group",
		get = function(info) 
			if type(info)~="string" then info=info[#info] end
			local func="Get"..string.upper(string.sub(info,1,1))..string.sub(info,2)
			if self[func] then return self[func](self) 
			else return self[info] end
		end,
		set = function(info,val)
			if type(info)~="string" then info=info[#info] end
			local func="Set"..string.upper(string.sub(info,1,1))..string.sub(info,2)
			if self[func] then self[func](self,val)
			else self[info]=val end
		end,
		args = {
			name = {
				type = "input",
				name = L["Name"],
				order = 10,
				width = "full"
			},
			select = {
				type = "toggle",
				name = L["Select"],
				width = "full",
				disabled = (self.scene==nil),
				order = 20,
				get =	function(_)
							if not self.scene then return false end
							if self.scene:GetSelectedMesh()==self then return true
							else return false end
						end,
				set =	function(_,val)
							if not self.scene then return end
							if val then self.scene:SetSelectedMesh(self)
							else self.scene:SetSelectedMesh(nil) end
						end
			},
			remove = {
				type = "execute",
				name = L["Remove"],
				width = "full",
				order = 30,
				func = function() self:Remove() end
			},
			color = {
				get = 	function(info)
							return self.r,self.g,self.b,self.a
						end,
				set =	function(info,r,g,b,a)
							self.r,self.g,self.b,self.a=r,g,b,a
						end,
				type = "color",
				name = L["Color"],
				hasAlpha = true,
				order = 40
			},
			follow = {
				type = "group",
				name = L["Follow behavior"],
				inline = true,
				order = 50,
				args = {
					followPlayer = {
						type = "toggle",
						name = L["Follow player"],
						order = 10,
					},
					followRotation = {
						type = "toggle",
						name = L["Follow rotation"],
						order = 20,
					},
					followUnit = {
						type = "input",
						name = L["Follow unit"],
						order = 30,
						set = 	function(info,val)
									val=string.gsub(string.gsub(val,"^%s*",""),"%s*$","")
									if string.len(val)==0 then val=nil end
									self.followUnit=val
								end
					},
					header1 = {
						type = "header",
						name = "",
						order = 40,
					},
					attach = {
						type = "execute",
						name = L["Attach"],
						order = 50,
						func = function() self:Attach() end
					},
					detach = {
						type = "execute",
						name = L["Detach"],
						order = 60,
						func = function() self:Detach() end
					},
					attachRotation = {
						type = "execute",
						name = L["Attach rotation"],
						order = 70,
						func = function() self:AttachRotation() end
					},
					detachRotation = {
						type = "execute",
						name = L["Detach rotation"],
						order = 80,
						func = function() self:DetachRotation() end
					},
				}
			},
			meshDeform = {
				type = "group",
				name = L["Mesh deform"],
				inline = true,
				order = 100,
				args = {
					drag = {
						type = "execute",
						name = "Drag",
						order = 10,
						width = "full",
						func = function() Core.mousePaint:EnableDrag(true,self) end
					},					
					meshTranslateX = {
						type = "range",
						name = L["X Position"],
						order = 20,
						width = "full",
						min = -500, max = 500, bigStep = 0.5,
						set = function(_,val) self.meshTranslateX=val ; self.vertices=nil end
					},
					meshTranslateY = {
						type = "range",
						name = L["Y Position"],
						order = 30,
						width = "full",
						min = -500, max = 500, bigStep = 0.5,
						set = function(_,val) self.meshTranslateY=val ; self.vertices=nil end
					},
					meshTranslateZ = {
						type = "range",
						name = L["Z Position"],
						order = 40,
						width = "full",
						min = -500, max = 500, bigStep = 0.5,
						set = function(_,val) self.meshTranslateZ=val ; self.vertices=nil end
					},
					meshScaleX = {
						type = "range",
						name = L["X Scale"],
						order = 50,
						width = "full",
						min = -100, max = 100, bigStep = 0.1,
						set = function(_,val) self.meshScaleX=val ; self.vertices=nil end
					},
					meshScaleY = {
						type = "range",
						name = L["Y Scale"],
						order = 60,
						width = "full",
						min = -100, max = 100, bigStep = 0.1,
						set = function(_,val) self.meshScaleY=val ; self.vertices=nil end
					},
					meshScaleZ = {
						type = "range",
						name = L["Z Scale"],
						order = 70,
						width = "full",
						min = -100, max = 100, bigStep = 0.1,
						set = function(_,val) self.meshScaleZ=val ; self.vertices=nil end
					},
					meshRotateZ = {
						type = "range",
						name = L["Z Rotate"],
						order = 80,
						width = "full",
						min = -360, max = 360, bigStep = 0.1,
						set = function(_,val) self.meshRotateZ=val*pi/180 ; self.vertices=nil end,
						get = function(_) return self.meshRotateZ*180/pi end
					}
				}
			},
		}
	}
end

function AVRMesh:GenerateMesh()
	ca=cos(self.meshRotateZ)
	sa=sin(self.meshRotateZ)
	local preA,preB,preC,preD=ca*self.meshScaleX,-sa*self.meshScaleY,
								sa*self.meshScaleX, ca*self.meshScaleY
	local v
	for i=1,#self.vertices do
		v=self.vertices[i]
		v[1],v[2],v[3]=	preA*v[1]+preB*v[2]+self.meshTranslateX,
						preC*v[1]+preD*v[2]+self.meshTranslateY,
						self.meshScaleZ*v[3]+self.meshTranslateZ
	end
end

function AVRMesh:Duplicate()
	local packed=self:Pack()
	local ret=AVR.meshGeneratorClasses[self.class]:New()
	ret:Unpack(packed)
	return ret
end

function AVRMesh:Pack()
	local s={
		n=self.name,
		c=self.class,
		tx=packDouble(self.meshTranslateX),
		ty=packDouble(self.meshTranslateY),
		tz=packDouble(self.meshTranslateZ),
		rz=packDouble(self.meshRotateZ),
		sx=packDouble(self.meshScaleX),
		sy=packDouble(self.meshScaleY),
		sz=packDouble(self.meshScaleZ),
		v=self.visible,
		r=packDouble(self.r),
		g=packDouble(self.g),
		b=packDouble(self.b),
		a=packDouble(self.a),
		fu=self.followUnit,
		fp=self.followPlayer,
		fr=self.followRotation
	}
	return s
end

function AVRMesh:Unpack(s)
	self.name=s.n or L["unnamed"]
	self.meshTranslateX=unpackDouble(s.tx) or 0.0
	self.meshTranslateY=unpackDouble(s.ty) or 0.0
	self.meshTranslateZ=unpackDouble(s.tz) or 0.0
	self.meshRotateZ=unpackDouble(s.rz) or 0.0
	self.meshScaleX=unpackDouble(s.sx) or 1.0
	self.meshScaleY=unpackDouble(s.sy) or 1.0
	self.meshScaleZ=unpackDouble(s.sz) or 1.0
	self.visible=s.v or true
	self.r=unpackDouble(s.r) or 1.0
	self.g=unpackDouble(s.g) or 1.0
	self.b=unpackDouble(s.b) or 1.0
	self.a=unpackDouble(s.a) or 1.0
	self.followUnit=s.fu
	self.followPlayer=s.fp or false
	self.followRotation=s.fr or false
	self.vertices={}
	self.lines={}
	self.triangles={}
end

function AVRMesh:SetScene(scene)
	self.scene=scene
	if self.scene then self.threed=self.scene.threed
	else self.threed=nil end
end

function AVRMesh:OnRemove()
	self.removed=true
end

function AVRMesh:Remove()
	if self.scene then self.scene:RemoveMesh(self) end
end

function AVRMesh:SelectMesh()
	if self.scene then self.scene:SetSelectedMesh(self) end
end

function AVRMesh:SetColor(r,g,b,a)
	if r~=nil then self.r=r end
	if g~=nil then self.g=g end
	if b~=nil then self.b=b end
	if a~=nil then self.a=a end
	return self
end

function AVRMesh:TranslateMesh(x,y,z)
	if x~=nil then self.meshTranslateX=self.meshTranslateX+x end
	if y~=nil then self.meshTranslateY=self.meshTranslateY+y end
	if z~=nil then self.meshTranslateZ=self.meshTranslateZ+z end
	self.vertices=nil
	return self
end

function AVRMesh:RotateMesh(z)
	if z~=nil then self.meshRotateZ=self.meshRotateZ+z end
	self.vertices=nil
	return self
end

function AVRMesh:ScaleMesh(x,y,z)
	if y==nil then
		y=x
		z=x
	elseif z==nil then
		z=1.0
	end
	self.meshScaleX=self.meshScaleX*x
	self.meshScaleY=self.meshScaleY*y
	self.meshScaleZ=self.meshScaleZ*z
	self.vertices=nil
end

function AVRMesh:SetName(name)
	self.name=name
	Core:OptionsChanged()
	return self
end

function AVRMesh:Attach(threed)
	if self.followUnit~=nil then
		self.followUnit=nil
		self.followPlayer=true
	elseif not self.followPlayer then
		threed=threed or self.threed
		self.followPlayer=true
		self.meshTranslateX,self.meshTranslateY,self.meshTranslateZ=
			self.meshTranslateX-threed.playerPosX,
			self.meshTranslateY-threed.playerPosY,
			self.meshTranslateZ-threed.playerPosZ
		self.vertices=nil
	end
	return self
end
function AVRMesh:Detach(threed)
	if self.followPlayer then
		threed=threed or self.threed
		self.followPlayer=false
		self.meshTranslateX,self.meshTranslateY,self.meshTranslateZ=
			self.meshTranslateX+threed.playerPosX,
			self.meshTranslateY+threed.playerPosY,
			self.meshTranslateZ+threed.playerPosZ		
		self.vertices=nil
	end
	return self
end
function AVRMesh:AttachRotation(threed)
	if not self.followRotation then
		threed=threed or self.threed
		self.followRotation=true
		self.meshRotateZ=self.meshRotateZ-threed.playerDirection
		self.vertices=nil
	end
	return self
end
function AVRMesh:DetachRotation(threed)
	if self.followRotation then
		threed=threed or self.threed
		self.followRotation=false
		self.meshRotateZ=self.meshRotateZ+threed.playerDirection
		self.vertices=nil
	end
	return self
end


function AVRMesh:SetFollowPlayer(value)
	if value==nil then value=true end
	self.followPlayer=value
	return self
end
function AVRMesh:SetFollowUnit(value)
	if value==nil then value=true end
	if self.followPlayer then
		self:SetFollowPlayer(false)
	end
	self.followUnit=value
	return self
end
function AVRMesh:SetFollowRotation(value)
	if value==nil then value=true end
	self.followRotation=value
	return self
end

function AVRMesh:OnUpdate(threed)
	if self.vertices==nil or self.lines==nil or self.triangles==nil then
		self.vertices={}
		self.lines={}
		self.triangles={}
		self:GenerateMesh()
	end

	if self.followRotation then
		self.rotateZ=threed.playerDirection
	else
		self.rotateZ=0.0
	end
	
	local ux,uy
	self.visible=true
	if self.followPlayer then
		self.translateX,self.translateY,self.translateZ=threed.playerPosX,threed.playerPosY,threed.playerPosZ
	elseif self.followUnit~=nil then
		ux,uy=threed:GetUnitPosition(self.followUnit)
		if ux==0.0 then
			self.visible=false
		else
			self.translateX,self.translateY,self.translateZ=ux,uy,threed.playerPosZ
		end
	else
		self.translateX,self.translateY,self.translateZ=0.0,0.0,0.0
	end
	self.scaleX,self.scaleY,self.scaleZ=1.0,1.0,1.0
end

function AVRMesh:AddVertex(x,y,z)
	insert(self.vertices,{x,y,z,0,0,0,0,0})
	return #self.vertices
end

function AVRMesh:FindVertex(x,y,z)
	for i,v in ipairs(self.vertices) do
		if v[1]==x and v[2]==y and v[3]==z then return i end
	end
	return nil
end

function AVRMesh:AddOrFindVertex(x,y,z)
	return self:FindVertex(x,y,z) or self:AddVertex(x,y,z)
end

function AVRMesh:AddLine(line,y1,z1,x2,y2,z2,width,a,r,g,b)
	if y1~=nil then
		local v1=self:AddOrFindVertex(line,y1,z1)
		local v2=self:AddOrFindVertex(x2,y2,z2)
		insert(self.lines,AVRLine:New(v1,v2,width,a,r,g,b))
	else
		insert(self.lines,line)
	end
end

function AVRMesh:AddTriangle(triangle,y1,z1,x2,y2,z2,x3,y3,z3,a,r,g,b)
	if y1~=nil then
		local v1=self:AddOrFindVertex(triangle,y1,z1)
		local v2=self:AddOrFindVertex(x2,y2,z2)
		local v3=self:AddOrFindVertex(x3,y3,z3)
		insert(self.triangles,AVRTriangle:New(v1,v2,v3,a,r,g,b))
	else
		insert(self.triangles,triangle)
	end
end

function AVRMesh:AddLineData(data,scale)
	scale=scale or 1.0
	for i=1,#data do
		self:AddLine(data[i][1]*scale,data[i][2]*scale,data[i][3]*scale,
					 data[i][4]*scale,data[i][5]*scale,data[i][6]*scale,
					 data[i][7],data[i][8],data[i][9],data[i][10],data[i][11])
	end
end

AVRLine = {Embed = Core.Embed}

function AVRLine:New(v1,v2,width,a,r,g,b)
	if self ~= AVRLine then return end
	local s={}
		
	AVRLine:Embed(s)
	
	s.r=r
	s.g=g
	s.b=b
	s.a=a
	s.v1=v1
	s.v2=v2
	s.width=(width==nil and 32 or width)
	s.visible=true

	return s
end

function AVRLine:Pack(bs)
	bs:WriteByte((self.r and 1 or 0)+(self.g and 2 or 0)+(self.b and 4 or 0)+(self.a and 8 or 0)+
				 (self.visible and 16 or 0))
	if self.r~=nil then bs:WriteByte(math.floor(self.r*255)) end
	if self.g~=nil then bs:WriteByte(math.floor(self.g*255)) end
	if self.b~=nil then bs:WriteByte(math.floor(self.b*255)) end
	if self.a~=nil then bs:WriteByte(math.floor(self.a*255)) end
	bs:WriteVariableBytes(self.v1)
	bs:WriteVariableBytes(self.v2)
	bs:WriteByte(math.floor(self.width))
end

function AVRLine:Unpack(bs)
	local mask=bs:ReadByte()
	if bit.band(mask,1)~=0 then self.r=bs:ReadByte()/255
	else self.r=nil end
	if bit.band(mask,2)~=0 then self.g=bs:ReadByte()/255
	else self.g=nil end
	if bit.band(mask,4)~=0 then self.b=bs:ReadByte()/255
	else self.b=nil end
	if bit.band(mask,8)~=0 then self.a=bs:ReadByte()/255
	else self.a=nil end
	if bit.band(mask,16)~=0 then self.visible=true
	else self.visible=false end
	self.v1=bs:ReadVariableBytes()
	self.v2=bs:ReadVariableBytes()
	self.width=bs:ReadByte()
end

AVRTriangle = {Embed = Core.Embed}

function AVRTriangle:New(v1,v2,v3,a,r,g,b)
	if self ~= AVRTriangle then return end
	local s={}
		
	AVRTriangle:Embed(s)
	
	s.r=r
	s.g=g
	s.b=b
	s.a=a
	s.v1=v1
	s.v2=v2
	s.v3=v3
	s.visible=true

	return s
end

function AVRTriangle:Pack(bs)
	bs:WriteByte((self.r and 1 or 0)+(self.g and 2 or 0)+(self.b and 4 or 0)+(self.a and 8 or 0)+
				 (self.visible and 16 or 0))
	if self.r~=nil then bs:WriteByte(math.floor(self.r*255)) end
	if self.g~=nil then bs:WriteByte(math.floor(self.g*255)) end
	if self.b~=nil then bs:WriteByte(math.floor(self.b*255)) end
	if self.a~=nil then bs:WriteByte(math.floor(self.a*255)) end
	bs:WriteVariableBytes(self.v1)
	bs:WriteVariableBytes(self.v2)
	bs:WriteVariableBytes(self.v3)
end

function AVRTriangle:Unpack(bs)
	local mask=bs:ReadByte()
	if bit.band(mask,1)~=0 then self.r=bs:ReadByte()/255
	else self.r=nil end
	if bit.band(mask,2)~=0 then self.g=bs:ReadByte()/255
	else self.g=nil end
	if bit.band(mask,4)~=0 then self.b=bs:ReadByte()/255
	else self.b=nil end
	if bit.band(mask,8)~=0 then self.a=bs:ReadByte()/255
	else self.a=nil end
	self.v1=bs:ReadVariableBytes()
	self.v2=bs:ReadVariableBytes()
	self.v3=bs:ReadVariableBytes()
	if bs:ReadByte()~=0 then self.visible=true else self.visible=false end
end

-------------------------------------------------------------------------
------     Mesh generators     ------------------------------------------
-------------------------------------------------------------------------

AVRDataMesh={Embed=Core.Embed}
AVRDataMesh.class="AVRDataMesh"
function AVRDataMesh:New()
	if self ~= AVRDataMesh then return end
	local s=AVRMesh:New()
	AVRDataMesh:Embed(s)
	s.class=AVRDataMesh.class
	s.vertexData={}
	s.lineData={}
	s.triangleData={}
	return s
end

function AVRDataMesh:CaptureMesh()
	self.vertexData=self.vertices
	self.lineData=self.lines
	self.triangleData=self.triangles
	self.vertices=nil
	self.lines=nil
	self.triangles=nil
end

function AVRDataMesh:GenerateMesh()
	local v
	for i=1,#self.vertexData do
		v=self.vertexData[i]
		insert(self.vertices,{v[1],v[2],v[3],0,0,0,0,0})
	end
	self.lines=self.lineData
	self.triangles=self.triangleData
	AVRMesh.GenerateMesh(self)
end

function AVRDataMesh:AddDataVertex(x,y,z)
	insert(self.vertexData,{x,y,z,0,0,0,0,0})
	return #self.vertexData
end

function AVRDataMesh:FindDataVertex(x,y,z)
	for i,v in ipairs(self.vertexData) do
		if v[1]==x and v[2]==y and v[3]==z then return i end
	end
	return nil
end

function AVRDataMesh:AddOrFindDataVertex(x,y,z)
	return self:FindDataVertex(x,y,z) or self:AddDataVertex(x,y,z)
end

function AVRDataMesh:AddDataLine(line,y1,z1,x2,y2,z2,width,a,r,g,b)
	if y1~=nil then
		local v1=self:AddOrFindDataVertex(line,y1,z1)
		local v2=self:AddOrFindDataVertex(x2,y2,z2)
		insert(self.lines,AVRLine:New(v1,v2,width,a,r,g,b))
	else
		insert(self.lines,line)
	end
	self.vertices=nil
end

function AVRDataMesh:AddDataTriangle(triangle,y1,z1,x2,y2,z2,x3,y3,z3,a,r,g,b)
	if y1~=nil then
		local v1=self:AddOrFindDataVertex(triangle,y1,z1)
		local v2=self:AddOrFindDataVertex(x2,y2,z2)
		local v3=self:AddOrFindDataVertex(x3,y3,z3)
		insert(self.triangleData,AVRTriangle:New(v1,v2,v3,a,r,g,b))
	else
		insert(self.triangleData,triangle)
	end
	self.vertices=nil
end

function AVRDataMesh:Pack()
	local s=AVRMesh.Pack(self)
	local bs=AVRByteStream:New()
	local v
	for i=1,#self.vertexData do
		v=self.vertexData[i]
		bs:WriteHalf(v[1])
		bs:WriteHalf(v[2])
		bs:WriteHalf(v[3])
	end
	bs:FlushBuffer()
	s.vd=bs.data
	
	bs=AVRByteStream:New()
	for i=1,#self.lineData do
		self.lineData[i]:Pack(bs)
	end
	bs:FlushBuffer()
	s.ld=bs.data

	bs=AVRByteStream:New()
	for i=1,#self.triangleData do
		self.triangleData[i]:Pack(bs)
	end
	bs:FlushBuffer()
	s.td=bs.data
	return s
end

function AVRDataMesh:Unpack(s)
	AVRMesh.Unpack(self,s)
	self.vertexData={}
	self.lineData={}
	self.triangleData={}
	
	local bs=AVRByteStream:New({data=s.vd})
	while not bs:IsEOS() do
		local v={0,0,0,0,0,0,0,0}
		v[1]=bs:ReadHalf()
		v[2]=bs:ReadHalf()
		v[3]=bs:ReadHalf()
		insert(self.vertexData,v)
	end
	
	bs=AVRByteStream:New({data=s.ld})
	while not bs:IsEOS() do
		local v=AVRLine:New()
		v:Unpack(bs)
		insert(self.lineData,v)
	end

	bs=AVRByteStream:New({data=s.td})
	while not bs:IsEOS() do
		local v=AVRTriangle:New()
		v:Unpack(bs)
		insert(self.triangleData,v)
	end
	self.vertices=nil
end



AVRCircleMesh={Embed=Core.Embed}
AVRCircleMesh.class="AVRCircleMesh"
function AVRCircleMesh:New(radius,segments,dashed,width)
	if self ~= AVRCircleMesh then return end
	local s=AVRMesh:New()
	AVRCircleMesh:Embed(s)
	s.class=AVRCircleMesh.class
	s.radius=radius or 10
	s.segments=segments or 30
	s.dashed=dashed or false	
	s.lineWidth=width
	s.vertices=nil
	s.name=L["Circle"]
	return s
end

function AVRCircleMesh:SetRadius(radius)
	self.radius=radius
	self.vertices=nil
	return self
end
function AVRCircleMesh:SetSegments(segments)
	self.segments=math.floor(segments)
	self.vertices=nil
	return self
end
function AVRCircleMesh:SetDashed(dashed)
	self.dashed=dashed
	self.vertices=nil
	return self
end

function AVRCircleMesh:GetOptions()
	local o=AVRMesh.GetOptions(self)
	o.args.circle = {
		type = "group",
		name = L["Circle properties"],
		inline = true,
		order = 80,
		args = {
			radius = {
				type = "range",
				name = L["Radius"],
				order = 10,
				width = "full",
				min = 0.0, max=500, bigStep=1.0
			},					
			segments = {
				type = "range",
				name = L["Segments"],
				order = 20,
				width = "full",
				min = 3, max=200, step=1, bigStep=1
			},					
			dashed = {
				type = "toggle",
				name = L["Dashed"],
				order = 30,
				width = "full",
			}
		}
	}
	return o
end

function AVRCircleMesh:GenerateMesh()
	local a
	local ox=self.radius
	local oy=0
	local x,y
	
	for i=1,self.segments do
		a=pi*2/self.segments*i
		
		x=cos(a)*self.radius
		y=sin(a)*self.radius
		
		if not self.dashed or (i%2)==0 then
			self:AddLine(ox,oy,0,x,y,0,self.lineWidth)
		end
		
		ox,oy=x,y
	end
	
	AVRMesh.GenerateMesh(self)
end

function AVRCircleMesh:Pack()
	local s=AVRMesh.Pack(self)
	s.rad=packDouble(self.radius)
	s.seg=self.segments
	s.das=self.dashed
	s.lw=self.lineWidth
	return s
end

function AVRCircleMesh:Unpack(s)
	AVRMesh.Unpack(self,s)
	self.radius=unpackDouble(s.rad) or 10
	self.segments=s.seg or 30
	if s.das~=nil then self.dashed=s.das
	else self.dashed=false end
	self.lw=s.lw
	self.vertices=nil
end



AVRFilledCircleMesh={Embed=Core.Embed}
AVRFilledCircleMesh.class="AVRFilledCircleMesh"
function AVRFilledCircleMesh:New(radius,segments)
	if self ~= AVRFilledCircleMesh then return end
	local s=AVRMesh:New()
	AVRFilledCircleMesh:Embed(s)
	s.class=AVRFilledCircleMesh.class
	s.radius=radius or 10
	s.segments=segments or 30
	s.vertices=nil
	s.name=L["Filled circle"]
	s.a=0.25
	return s
end

function AVRFilledCircleMesh:Pack()
	local s=AVRMesh.Pack(self)
	s.rad=packDouble(self.radius)
	s.seg=self.segments
	return s
end

function AVRFilledCircleMesh:Unpack(s)
	AVRMesh.Unpack(self,s)
	self.radius=unpackDouble(s.rad) or 10
	self.segments=s.seg or 30
	self.vertices=nil
end

function AVRFilledCircleMesh:SetRadius(radius)
	self.radius=radius
	self.vertices=nil
	return self
end
function AVRFilledCircleMesh:SetSegments(segments)
	self.segments=math.floor(segments)
	self.vertices=nil
	return self
end

function AVRFilledCircleMesh:GetOptions()
	local o=AVRMesh.GetOptions(self)
	o.args.filledCircle = {
		type = "group",
		name = L["Circle properties"],
		inline = true,
		order = 80,
		args = {
			radius = {
				type = "range",
				name = L["Radius"],
				order = 10,
				width = "full",
				min = 0.0, max=500, bigStep=1.0
			},					
			segments = {
				type = "range",
				name = L["Segments"],
				order = 20,
				width = "full",
				min = 3, max=200, step=1, bigStep=1
			}
		}
	}
	return o
end


function AVRFilledCircleMesh:GenerateMesh()
	local a
	local ox=self.radius
	local oy=0
	local x,y
	
	for i=1,self.segments do
		a=pi*2/self.segments*i
		
		x=cos(a)*self.radius
		y=sin(a)*self.radius
		
		self:AddTriangle(0,0,0,ox,oy,0,x,y,0)
		
		ox,oy=x,y
	end
	
	AVRMesh.GenerateMesh(self)
end



AVRConeMesh={Embed=Core.Embed}
AVRConeMesh.class="AVRConeMesh"
function AVRConeMesh:New(radius,angle,segments,dashed,circle,width)
	if self ~= AVRConeMesh then return end
	local s=AVRMesh:New()
	AVRConeMesh:Embed(s)
	s.class=AVRConeMesh.class
	s.radius=radius or 10
	s.angle=angle or 30
	s.angleR=s.angle*pi/180
	s.segments=segments or max(3,math.ceil(s.angleR/(2*pi)*30))
	s.dashed=dashed or false	
	if circle~=nil then s.circle=circle
	else s.circle=true end
	s.lineWidth=width
	s.vertices=nil
	s.name=L["Cone"]
	return s
end

function AVRConeMesh:Pack()
	local s=AVRMesh.Pack(self)
	s.rad=packDouble(self.radius)
	s.ang=packDouble(self.angle)
	s.seg=self.segments
	s.das=self.dashed
	s.cir=circle
	s.lw=self.lineWidth
	return s
end

function AVRConeMesh:Unpack(s)
	AVRMesh.Unpack(self,s)
	self.radius=unpackDouble(s.rad) or 10
	self.angle=unpackDouble(s.ang) or 30
	self.angleR=self.angle*pi/180
	self.segments=s.seg or max(3,math.ceil(self.angleR/(2*pi)*30))
	self.dashed=s.das or false
	self.lineWidth=s.lw
	if s.cir then self.circle=s.cir
	else self.circle=true end
	self.vertices=nil
end

function AVRConeMesh:SetRadius(radius)
	self.radius=radius
	self.vertices=nil
	return self
end
function AVRConeMesh:SetAngle(angle)
	self.angle=angle
	self.angleR=self.angle*pi/180
	self.vertices=nil
	return self
end
function AVRConeMesh:SetSegments(segments)
	self.segments=math.floor(segments)
	self.vertices=nil
	return self
end
function AVRConeMesh:SetDashed(dashed)
	self.dashed=dashed
	self.vertices=nil
	return self
end
function AVRConeMesh:SetCircle(circle)
	self.circle=circle
	self.vertices=nil
	return self
end

function AVRConeMesh:GetOptions()
	local o=AVRMesh.GetOptions(self)
	o.args.cone = {
		type = "group",
		name = L["Cone properties"],
		inline = true,
		order = 80,
		args = {
			radius = {
				type = "range",
				name = L["Radius"],
				order = 10,
				width = "full",
				min = 0.0, max=500, bigStep=1.0
			},					
			angle = {
				type = "range",
				name = L["Angle"],
				order = 15,
				width = "full",
				min = 0.0, max=360, bigStep=1.0
			},					
			segments = {
				type = "range",
				name = L["Segments"],
				order = 20,
				width = "full",
				min = 3, max=200, step=1, bigStep=1
			},
			dashed = {
				type = "toggle",
				name = L["Dashed"],
				order = 30,
				width = "full"
			},
			circle = {
				type = "toggle",
				name = L["Circle"],
				order = 40,
				width = "full"
			}
		}
	}
	return o
end

function AVRConeMesh:GenerateMesh()

	local a
	local ox=cos(pi/2-self.angleR/2)*self.radius
	local oy=sin(pi/2-self.angleR/2)*self.radius
	local ex=cos(pi/2+self.angleR/2)*self.radius
	local ey=sin(pi/2+self.angleR/2)*self.radius
	local x,y
	
	self:AddLine(0,0,0,ox,oy,0,self.lineWidth)
	self:AddLine(0,0,0,ex,ey,0,self.lineWidth)
	
	if self.circle then
		for i=1,self.segments do
			a=pi/2-self.angleR/2+self.angleR/self.segments*i
			
			x=cos(a)*self.radius
			y=sin(a)*self.radius
			
			if not self.dashed or (i%2)==0 then
				self:AddLine(ox,oy,0,x,y,0,self.lineWidth)
			end
			
			ox,oy=x,y
		end
	end
	
	AVRMesh.GenerateMesh(self)
end



AVRYardStickMesh={Embed=Core.Embed}
AVRYardStickMesh.class="AVRYardStickMesh"
function AVRYardStickMesh:New(max,min,vertical,divisions)
	if self ~= AVRYardStickMesh then return end
	local s=AVRMesh:New()
	AVRYardStickMesh:Embed(s)
	s.class=AVRYardStickMesh.class
	s.min=min and math.floor(min) or 0
	s.max=max and math.ceil(max) or 10
	s.divisions=divisions or 2
	if vertical~=nil then s.vertical=vertical
	else s.vertical=false end
	s.vertices=nil
	s.followPlayer=true
	s.followRotation=true
	s.name=L["Yard stick"]
	return s
end

function AVRYardStickMesh:Pack()
	local s=AVRMesh.Pack(self)
	s.min=self.min
	s.max=self.max
	s.ver=self.vertical
	s.div=self.divisions
	return s
end

function AVRYardStickMesh:Unpack(s)
	AVRMesh.Unpack(self,s)
	self.min=s.min or 0
	self.max=s.max or 10
	self.vertical=s.ver or false
	self.divisions=s.div or 2
	self.vertices=nil
end


function AVRYardStickMesh:SetMin(min)
	self.min=math.floor(min)
	self.vertices=nil
	return self
end
function AVRYardStickMesh:SetMax(max)
	self.max=math.ceil(max)
	self.vertices=nil
	return self
end
function AVRYardStickMesh:SetDivisions(divisions)
	self.divisions=math.floor(divisions)
	self.vertices=nil
	return self
end
function AVRYardStickMesh:SetVertical(vertical)
	self.vertical=vertical
	self.vertices=nil
	return self
end


function AVRYardStickMesh:GetOptions()
	local o=AVRMesh.GetOptions(self)
	o.args.yardStick = {
		type = "group",
		name = L["Yard stick properties"],
		inline = true,
		order = 80,
		args = {
			min= {
				type = "range",
				name = L["Min"],
				order = 10,
				width = "full",
				min = -100, max=100, step=1.0, bigStep=1.0
			},					
			max= {
				type = "range",
				name = L["Max"],
				order = 20,
				width = "full",
				min = -100, max=100, step=1.0, bigStep=1.0
			},
			vertical = {
				type = "toggle",
				name = L["Vertical"],
				order = 30,
				width = "full"
			},
			divisions = {
				name = L["Divisions"],
				type = "range",
				order = 40,
				min = 1,
				max = 100,
				bigStep = 1
			},
		}
	}
	return o
end
function AVRYardStickMesh:GenerateMesh()
	local s=0
	local w=0
	local h=0
	while s<=(self.max-self.min)*self.divisions do
		h=self.min+1/self.divisions*s
		if (s%self.divisions)==0 then w=0.2
		elseif (s%(self.divisions/2))==0 then w=0.1
		elseif (s%(self.divisions/4))==0 then w=0.085
		elseif (s%(self.divisions/10))==0 then w=0.07
		else w=0.05 end
	
		if self.vertical then self:AddLine(-w,0,h, w,0,h, 0)
		else self:AddLine(-w,h,0, w,h,0, 0) end
		s=s+1
	end
	if self.vertical then self:AddLine(0,0,self.min, 0,0,self.max, 0)
	else self:AddLine(0,self.min,0, 0,self.max,0, 0) end
	
	AVRMesh.GenerateMesh(self)
end

AVRMarkerMesh={Embed=Core.Embed}
AVRMarkerMesh.class="AVRMarkerMesh"
function AVRMarkerMesh:New(radius,spokes)
	if self ~= AVRMarkerMesh then return end
	local s=AVRMesh:New()
	AVRMarkerMesh:Embed(s)
	s.class=AVRMarkerMesh.class
	s.radius=radius or 1
	s.spokes=spokes or 6
	s.vertices=nil
	s.name="Marker"
	return s
end

function AVRMarkerMesh:SetRadius(radius)
	self.radius=radius
	self.vertices=nil
end

function AVRMarkerMesh:SetSpokes(spokes)
	self.spokes=spokes
	self.vertices=nil
end

function AVRMarkerMesh:Pack()
	local s=AVRMesh.Pack(self)
	s.rad=self.radius
	s.spo=self.spokes
	return s
end

function AVRMarkerMesh:Unpack(s)
	AVRMesh.Unpack(self,s)
	self.radius=s.rad or 1
	self.spokes=s.spo or 6
	self.vertices=nil
end

function AVRMarkerMesh:GenerateMesh()
	local a
	local ox=self.radius
	local oy=0
	local x,y
	
	for i=1,self.spokes do
		a=pi*2/self.spokes*i
		
		x=cos(a)*self.radius
		y=sin(a)*self.radius
		
		if (i%2)==0 then
			self:AddTriangle(0,0,0,ox,oy,0,x,y,0,nil,0,0,0)
		else
			self:AddTriangle(0,0,0,ox,oy,0,x,y,0)
		end
		
		ox,oy=x,y
	end
	
	AVRMesh.GenerateMesh(self)
end


-------------------------------------------------------------------------
------     Predefined meshes     ----------------------------------------
-------------------------------------------------------------------------


AVRCompass={}
function AVRCompass:New()
	if self ~= AVRCompass then return end
	local mesh=AVRDataMesh:New()
	mesh:AddLineData({	
		{	0,		20,		0,		0,		-10,	0,		nil}, -- n - s
		{	-10,	0,		0,		10,		0,		0,		nil}, -- w - e
		{	-4,		-4,		0,		4,		4,		0,		nil}, -- sw - ne
		{	-4,		4,		0,		4,		-4,		0,		nil}, -- nw - se
		
		{	2,		8,		0,		2,		12,		0,		nil}, -- n
		{	2,		12,		0,		5,		8,		0,		nil},
		{	5,		8,		0,		5,		12,		0,		nil},

		{	5,		-7,		0,		2,		-7,		0,		nil}, -- s
		{	2,		-7,		0,		2,		-9,		0,		nil},
		{	2,		-9,		0,		5,		-9,		0,		nil},
		{	5,		-9,		0,		5,		-11,	0,		nil},
		{	5,		-11,	0,		2,		-11,	0,		nil},

		{	10,		2,		0,		8,		2,		0,		nil}, -- e
		{	8,		2,		0,		8,		6,		0,		nil},
		{	8,		6,		0,		10,		6,		0,		nil},
		{	8,		4,		0,		9,		4,		0,		nil},

		{	-13,	6,		0,		-11,	2,		0,		nil}, -- w
		{	-11,	2,		0,		-10,	4,		0,		nil},
		{	-10,	4,		0,		-9,		2,		0,		nil},
		{	-9,		2,		0,		-7,		6,		0,		nil}
	})
	mesh:CaptureMesh()
	mesh.followPlayer=true
	mesh.name=L["Compass"]
	return mesh
end

AVRVargoth={}
function AVRVargoth:New()
	if self ~= AVRVargoth then return end
	local mesh=AVRCircleMesh:New(0.5)
	mesh:TranslateMesh(-2.15,2.15,0)
	mesh.followPlayer=true
	mesh.followRotation=true
	mesh.name=L["Vargoth"]
	return mesh
end

AVRBlink={}
function AVRBlink:New()
	if self ~= AVRBlink then return end
	local mesh=AVRCircleMesh:New(0.5)
	mesh:TranslateMesh(0,20,0)
	mesh.followPlayer=true
	mesh.followRotation=true
	mesh.name=L["Blink"]
	return mesh
end

AVRPositionCrosshair={}
function AVRPositionCrosshair:New(scale)
	if self ~= AVRPositionCrosshair then return end
	local mesh=AVRDataMesh:New()
	scale=scale or 0.5
	mesh:AddLineData({	
		{	0,		5,		0,		0,		-5,		0,		0},
		{	5,		0,		0,		-5,		0,		0,		0}
	},scale)	
	mesh:CaptureMesh()
	mesh.followPlayer=true
	mesh.followRotation=true
	mesh.name=L["Position Crosshair"]
	return mesh
end

AVRUnitMarker={}
function AVRUnitMarker:New(unit)
	if self ~= AVRUnitMarker then return end
	local mesh=AVRMarkerMesh:New()
	mesh.followUnit=unit
	mesh.name=L["Unit marker"]
	mesh:SetColor(nil,nil,nil,0.25)
	return mesh
end

AVRTargetMarker={}
function AVRTargetMarker:New(height)
	if self ~= AVRTargetMarker then return end
	local mesh=AVRMarkerMesh:New()
	local oldUpdate=mesh.OnUpdate
	local ix,iy,px,py
	mesh.OnUpdate=function(threed)
		oldUpdate(self,threed)
		
		px,py=Core:GetTargetPlatePos()
		if px==nil then 
			self.visible=false
			return
		end
		self.visible=true
		threed:InvertCameraMatrix()
		ix,iy,_=threed:InverseProject(px,py,self.height-threed.playerHeight)
		self.translateX,self.translateY=self.translateX+ix,self.translateY+iy
	end
	mesh.unit=unit
	mesh.height=height or 2.5
	mesh.name=L["Target marker"]
	
	return mesh
end



-------------------------------------------------------------------------
------     Mesh options and registering     -----------------------------
-------------------------------------------------------------------------

AVR.meshGeneratorClasses={}
AVR.receiveMeshClasses={}
AVR.guiMeshTemplates={}

function AVR:RegisterMesh(meshClass,guiName,generator,receive)
	local template=nil
	if generator then self.meshGeneratorClasses[meshClass.class]=meshClass end
	if receive then self.receiveMeshClasses[meshClass.class]=meshClass end
	if guiName then
		if meshClass['NewTemplate'] then template=meshClass:NewTemplate()
		else template=meshClass:New() end
		self.guiMeshTemplates[guiName]=template
	end
end

function AVR:GetNewMeshArgs(scene)
	local ret={}
	for name,template in pairs(self.guiMeshTemplates) do
		ret[name]=template:GetOptions()
		ret[name].name=name
		ret[name].args.add={
			name=L["Add"],
			order=1,
			width="full",
			type="execute",
			func= 	function()
						local n=template:Duplicate()
						local mesh=scene:AddMesh(n)
						if not mesh.followPlayer and mesh.followUnit==nil then
							mesh.followPlayer=true
							mesh:Detach()
						end
					end
		}
		if ret[name].args.follow then
			ret[name].args.follow.args.header1=nil
			ret[name].args.follow.args.attach=nil
			ret[name].args.follow.args.attachRotation=nil
			ret[name].args.follow.args.detach=nil
			ret[name].args.follow.args.detachRotation=nil
		end
		ret[name].args.remove=nil
	end
	return ret
end

AVR:RegisterMesh(AVRDataMesh,nil,true,true)
AVR:RegisterMesh(AVRCircleMesh,L["Circle"],true,true)
AVR:RegisterMesh(AVRFilledCircleMesh,L["Filled circle"],true,true)
AVR:RegisterMesh(AVRConeMesh,L["Cone"],true,true)
AVR:RegisterMesh(AVRYardStickMesh,L["Yard stick"],true,true)
AVR:RegisterMesh(AVRMarkerMesh,nil,true,true)
AVR:RegisterMesh(AVRCompass,L["Compass"],false,false)
AVR:RegisterMesh(AVRVargoth,L["Vargoth"],false,false)
AVR:RegisterMesh(AVRBlink,L["Blink location"],false,false)
AVR:RegisterMesh(AVRPositionCrosshair,L["Position Crosshair"],false,false)
AVR:RegisterMesh(AVRUnitMarker,L["Unit marker"],false,false)
--AVR:RegisterMesh(AVRTargetMarker,L["Target marker"],false,false)



-------------------------------------------------------------------------
------     Effects     --------------------------------------------------
-------------------------------------------------------------------------

function AVRBreath(mesh,amount,speed)
	amount=amount or 0.1
	speed=speed or 2.0
	local original=mesh.OnUpdate
	mesh.OnUpdate=function(self,threed)
		original(self,threed)
		local s=1.0+sin(GetTime()*speed)*amount
		self.scaleX,self.scaleY,self.scaleZ=self.scaleX*s,self.scaleY*s,self.scaleZ*s
	end
	mesh.RemoveBreath=function(self)
		self.OnUpdate=original
		self.RemoveBreath=nil
	end
end

function AVRPulse(mesh,r,g,b,speed)
	r=r or 1.0
	g=g or 0.0
	b=b or 0.0
	local ro,go,bo=mesh.r,mesh.g,mesh.b
	speed=speed or 2.0
	local original=mesh.OnUpdate
	mesh.OnUpdate=function(self,threed)
		original(self,threed)
		local s=sin(GetTime()*speed)*0.5+0.5
		local t=1-s
		self.r,self.g,self.b=s*r+t*ro,s*g+t*go,s*b+t*bo
	end
	mesh.RemovePulse=function(self)
		self.r,self.g,self.b=ro,go,bo
		self.OnUpdate=original
		self.RemovePulse=nil
	end	
end
