local ADDON_NAME="AVR"
local L = LibStub("AceLocale-3.0"):GetLocale(ADDON_NAME, true)

local _G = _G
local pairs = pairs
local ipairs = ipairs
local remove = table.remove
local insert = table.insert
local max = math.max
local min = math.min
local abs = math.abs
local atan2 = math.atan2
local sin = math.sin
local cos = math.cos
local type = type
local select = select
local next = next
local GetTime = GetTime
local deg2rad = math.pi/180.0
local pi = math.pi
local format = string.format

local Core=AVR

AVRScene = {Embed = Core.Embed}
local T=AVRScene
T.class="AVRScene"
function T:New(threed)
	if self ~= T then return end
	local s={}
	
	T:Embed(s)

	s.threed=threed
	s.meshes={}
	s.selection=0
	s.class=T.class
	
	s.visible=true
	s.name=L["Scene"]
	s.from=nil
	s.id=0
	s.zone=nil
	s.shareChannel="Raid"
	s.shareTarget=""
	
	s.autoFireOptionChanges=true
	s.sceneManager=nil
	
	return s
end

function T:Pack()
	local s={
		name=self.name,
		class=self.class,
		visible=self.visible,
		from=self.from,
		id=self.id,
		selection=self.selection,
		zone=self.zone,
		shareChannel=self.shareChannel,
		shareTarget=self.shareTarget,
		meshes={}
	}
	for i=1,#self.meshes do
		insert(s.meshes,self.meshes[i]:Pack())
	end
	return s
end

function T:Unpack(s,commReceive)
	self.name=s.name or L["Scene"]
	self.visible=s.visible
	if self.visible==nil then self.visible=true end
	self.from=s.from
	self.id=s.id or 0
	self.selection=s.selection or 0
	self.zone=s.zone
	self.shareChannel=s.shareChannel or "RAID"
	self.shareTarget=s.shareTarget or ""
	self.meshes={}
	for i=1,#s.meshes do
		local class=s.meshes[i].c
		local m
		
		if AVR.meshGeneratorClasses[class] then
			if not commReceive or AVR.receiveMeshClasses[class] then
				m=AVR.meshGeneratorClasses[class]:New()
				m:Unpack(s.meshes[i])
				insert(self.meshes,m)
				m:SetScene(self)
			else
				Core:Print(format(L["Received mesh with class %s but that class is not receivable"],class))
			end
		else
			Core:Print(format(L["Trying to unpack mesh, unknown class %s"],class))
		end
	end
	
	if self.autoFireOptionChanges then Core:OptionsChanged() end
end


function T:GetToolSettings(info)
	if type(info)=="string" then
		info={info}
	end
	for i=1,#info do
		if self.toolSettings[info[i]] then
			local t=self.toolSettings
			for j=i,#info do
				t=t[info[j]]
				if t==nil then return nil end
			end
			return t
		end
	end
	return nil
end

function T:SetSceneManager(m)
	self.sceneManager=m
end

function T:SelectScene()
	if not self.sceneManager then return end
	self.sceneManager:SetSelectedScene(self)
end

function T:DrawScene()
	for i=1,#self.meshes do
		self.meshes[i]:OnUpdate(self.threed)
		self.threed:DrawMesh(self.meshes[i])
	end
end

function T:AddMesh(mesh,select,selectScene)
	mesh:SetScene(self)
	insert(self.meshes,mesh)
	if select==nil or select then
		self.selection=#self.meshes
	end
	if selectScene==nil or selectScene then
		self:SelectScene()
	end
	
	if self.autoFireOptionChanges then Core:OptionsChanged() end
	return mesh
end

function T:RemoveMesh(mesh)
	local sel=self:GetSelectedMesh()
	for i=1,#self.meshes do
		if self.meshes[i]==mesh then
			table.remove(self.meshes,i)
			mesh:OnRemove()
			break
		end
	end
	if sel then self:SetSelectedMesh(sel) end
	
	if self.autoFireOptionChanges then Core:OptionsChanged() end
end

function T:GetLastMesh()
	if #self.meshes==0 then return nil end
	return self.meshes[#self.meshes]
end

function T:SetSelectedMeshNum(num)
	if num==nil or num<0 or num>#self.meshes then num=0 end
	self.selection=num
end

function T:SetSelectedMesh(mesh)
	if mesh~=nil then
		for i=1,#self.meshes do
			if self.meshes[i]==mesh then
				self.selection=i
				if self.autoFireOptionChanges then Core:OptionsChanged() end
				return
			end
		end
	end
	self.selection=0
	if self.autoFireOptionChanges then Core:OptionsChanged() end
end

function T:GetSelectedMesh()
	if self.selection<=0 then return nil
	else return self.meshes[self.selection] end
end

function T:MakeOwnCopy()
	local p=self:Pack()
	local scene=AVRScene:New(self.threed)
	scene:Unpack(p)
	self.sceneManager:AddScene(scene)
end

function T:ClearScene()
	self.selection=0
	for i=1,#self.meshes do
		self.meshes[i]:OnRemove()
	end
	self.meshes={}
	
	if self.autoFireOptionChanges then Core:OptionsChanged() end
end

function T:OnRemove()
end

function T:Remove()
	self.sceneManager:RemoveScene(self)
end

function T:GetToolSettings(info,val)
	if type(info)=="string" then
		info={info}
	end
	for i=1,#info do
		if self.toolSettings[info[i]] then
			local t=self.toolSettings
			for j=i,#info-1 do
				t=t[info[j]]
				if t==nil then return end
			end
			t[info[#info]]=val
		end
	end
	return nil
end

function T:GetMeshesOptionsArgs()
	local ret={}
	local i,m
	for i,m in ipairs(self.meshes) do
		ret[""..i]=m:GetOptions()
	end
	return ret
end

function T:GetOptions()
	if not self.toolSettings then
		self.toolSettings={
			paint = {
				followPlayer = false,
				followRotation = false,
				detail = 0.5
			}
		}
	end

	local r={
		name = (self.from and (self.from..": "..self.name) or self.name),
		type = "group",
		get = function(info) 
			if type(info)~="string" then info=info[#info] end
			return self[info]
		end,
		set = function(info,val)
			if type(info)~="string" then info=info[#info] end
			self[info]=val
		end,
		args={
			name = {
				type = "input",
				name = L["Name"],
				order = 10,
				width = "full",
				set = 	function(_,val)
							self.name=val
							Core:OptionsChanged()
						end
			},
			visible = {
				type = "toggle",
				name = L["Visible"],
				order = 15
			},
			select = {
				type = "toggle",
				name = L["Select"],
				order = 20,
				get =	function(_)
							if Core.sceneManager:GetSelectedScene()==self then return true
							else return false end
						end,
				set =	function(_,val)
							if val then Core.sceneManager:SetSelectedScene(self)
							else Core.sceneManager:SetSelectedScene(nil) end
						end
			},
			clear = {
				type = "execute",
				name = L["Clear scene"],
				order = 25,
				width = "full",
				confirm = true,
				func = function() self:ClearScene() end
			},
			remove = {
				type = "execute",
				name = L["Remove scene"],
				order = 30,
				width = "full",
				confirm = true,
				func = function() self:Remove() end
			},
			zone = {
				name = L["Zone"],
				type = "input",
				order = 40,
				set = 	function(_,val)
							val=string.gsub(string.gsub(val,"^%s*",""),"%s*$","")
							if string.len(val)==0 then val=nil end
							self.zone=val
						end
				
			},
			currentZone = {
				name = L["Current zone"],
				type = "execute",
				order = 50,
				func = function() self.zone=Core.zoneData:GetCurrentZone() end
			},
			sharing = {
				type = "group",
				name = L["Share"],
				order = 60,
				inline = true,
				args = {
					shareChannel = {
						name = L["Channel"],
						type = "select",
						order = 10,
						values = {RAID="Raid",PARTY="Party",GUILD="Guild",BATTLEGROUND="Battleground",WHISPER="Whisper"},
						set = 	function(_,val)
									if val~="WHISPER" then self.shareTarget="" end
									self.shareChannel=val
								end
					},
					shareTarget = {
						name = L["Whisper target"],
						type = "input",
						order = 20,
						set = 	function(_,val)
									if self.shareChannel=="WHISPER" then self.shareTarget=val end
								end
					},
					send = {
						name = L["Send"],
						type = "execute",
						order = 30,
						func =	function()
									Core:SendScene(self,self.shareChannel,self.shareTarget)
								end
					}
				}
			},
			id = {
				name = L["Scene id"],
				type = "input",
				disabled = true,
				order = 70,
				set = function() end,
				get = function() return ''..self.id end
			},
			from = {
				name = L["Owner"],
				type = "input",
				disabled = true,
				order = 80,
				set = function() end,
			},
			ownCopy = {
				name = L["Make own copy"],
				type = "execute",
				order = 90,
				disabled = (self.from==nil),
				func = function() scene:MakeOwnCopy() end
			},
			paint = {
				type = "group",
				name = L["Paint"],
				order = 140,
				get = function(info) self:GetToolSettings(info) end,
				set = function(info,val) self:SetToolSettings(info,val) end,
				args = {
					followPlayer = { name=L["Follow player"], type = "toggle" },
					followRotation = { name=L["Follow rotation"], type = "toggle" },
					detail = { 
						name = L["Detail level"],
						type = "range",
						min = 0.1,
						max = 5.0,
						bigStep = 0.1,
					},
					start = { 
						type = "execute",
						name = L["Start"],
						func = function()
							Core.mousePaint:EnableDraw(true,self.toolSettings.paint.followPlayer,self.toolSettings.paint.followRotate,self)
							Core.mousePaint:SetDetailLevel(self.toolSettings.paint.detail)
						end
					}
				}
			},
			meshes = {
				type = "group",
				name = L["Meshes"],
				order = 150,
				args = self:GetMeshesOptionsArgs(self)
			},
			newMesh = {
				type = "group",
				name = L["Add new mesh"],
				order = 160,
				args = AVR:GetNewMeshArgs(self)
			},
		}
	}
	if not AVR.receiveSceneClasses[self.class] then r.args.sharing=nil end
	return r
end

AVR.sceneClasses={}
AVR.receiveSceneClasses={}
AVR.guiSceneClasses={}

function AVR:RegisterSceneClass(sceneClass,guiName,receive)
	self.sceneClasses[sceneClass.class]=sceneClass
	if receive then self.receiveSceneClasses[sceneClass.class]=sceneClass end
	if guiName then self.guiSceneClasses[sceneClass.class]=guiName end
end

AVR:RegisterSceneClass(AVRScene,L["Normal scene"],true)
